__all__ = [
    "__version__",
    "connect",
    "connect_local_config",
    "http_transport",
    "exasol_mapper",
    "ExaError",
    "ExaCommunicationError",
    "ExaConcurrencyError",
    "ExaRuntimeError",
    "ExaRequestError",
    "ExaAuthError",
    "ExaQueryError",
    "ExaQueryAbortError",
    "ExaQueryTimeoutError",
    "ExaConnection",
    "ExaConnectionError",
    "ExaConnectionDsnError",
    "ExaConnectionFailedError",
    "ExaStatement",
    "ExaFormatter",
    "ExaLogger",
    "ExaExtension",
    "ExaMetaData",
    "ExaHTTPTransportWrapper",
    "ExaLocalConfig",
    "ExaTimeDelta",
    "PROTOCOL_V1",
    "PROTOCOL_V2",
    "PROTOCOL_V3",
]

from .connection import ExaConnection
from .constant import (
    PROTOCOL_V1,
    PROTOCOL_V2,
    PROTOCOL_V3,
)
from .exceptions import (
    ExaAuthError,
    ExaCommunicationError,
    ExaConcurrencyError,
    ExaConnectionDsnError,
    ExaConnectionError,
    ExaConnectionFailedError,
    ExaError,
    ExaQueryAbortError,
    ExaQueryError,
    ExaQueryTimeoutError,
    ExaRequestError,
    ExaRuntimeError,
)
from .ext import ExaExtension
from .formatter import ExaFormatter
from .http_transport import ExaHTTPTransportWrapper
from .local_config import ExaLocalConfig
from .logger import ExaLogger
from .mapper import (
    ExaTimeDelta,
    exasol_mapper,
)
from .meta import ExaMetaData
from .statement import ExaStatement
from .version import __version__


def connect(**kwargs) -> ExaConnection:
    """
    Create a new connection object.

    Args:
        **kwargs:
            For details, refer to the :class:`pyexasol.ExaConnection` class.
    """
    return ExaConnection(**kwargs)


def connect_local_config(config_section, config_path=None, **kwargs) -> ExaConnection:
    """
    Constructor for connection objects based on a local config file.

    Info:
        - The default config path is ~/.pyexasol.ini
        - Extra arguments override values from config

    Args:

        config_section:
            Name of config section (required!)
        config_path:
            Custom path to local config file
        kwargs:
            Arguments for "connect()" function
    """

    conf = ExaLocalConfig(config_path)
    conf_args = conf.get_args(config_section)

    return connect(**{**conf_args, **kwargs})


def http_transport(
    ipaddr, port, compression=False, encryption=True
) -> ExaHTTPTransportWrapper:
    """
    Constructor for HTTP Transport wrapper for parallel HTTP Transport (EXPORT or IMPORT)

    Args:
        ipaddr:
            IP address of one of Exasol nodes received from :meth:`pyexasol.ExaConnection.get_nodes`
        port:
            Port of one of Exasol nodes received from :meth:`pyexasol.ExaConnection.get_nodes`
        compression:
            Use zlib compression for HTTP transport, must be the same as `compression` of main connection
        encryption:
            Use SSL/TLS encryption for HTTP transport, must be the same as `encryption` of main connection

    Info:
        Compression and encryption arguments should match :func:`pyexasol.connect`

        How to use:

        #. Parent process opens main connection to Exasol with pyexasol.connect()
        #. Parent process creates any number of child processes (possibly on remote host or another container)
        #. Every child process starts HTTP transport sub-connection with pyexasol.http_transport()
        #.  and gets "ipaddr:port" string using ExaHTTPTransportWrapper.address
        #. Every child process sends address string to parent process using any communication method (Pipe, Queue, Redis, etc.)
        #. Parent process runs .export_parallel() or .import_parallel(), which initiates EXPORT or IMPORT query in Exasol
        #. Every child process receives or sends a chunk of data using ExaHTTPTransportWrapper.export_*() or .import_*()
        #. Parent process waits for Exasol query and for child processes to finish

        All child processes should run in parallel.
        It is NOT possible to process some data first, and process some more data later.

        If an exception is raised in child process, it will close the pipe used for HTTP transport.
        Closing the pipe prematurely will cause SQL query to fail and will raise an exception in parent process.
        Parent process is responsible for closing other child processes and cleaning up.

        PyExasol does not provide a complete solution to manage child processes, only examples.
        The final solution depends on your hardware, network configuration, cloud provider and container orchestration software.
    """
    return ExaHTTPTransportWrapper(ipaddr, port, compression, encryption)



from . import constant
from .exceptions import ExaRuntimeError


class ExaMetaData:
    """
    This class implements lock-free metadata requests using ``/*snapshot execution*/``
    SQL hint, as described further in https://docs.exasol.com/db/latest/database_concepts/snapshot_mode.htm.

    Note:
        If you still get locks, please make sure to update Exasol server to the latest minor version.

    Examples:

        You may access these functions using `.meta` property of connection object.

        >>> C = pyexasol.connect(...)
        ... print(C.meta.sql_columns('SELECT 1 AS id'))
    """

    snapshot_execution_hint = "/*snapshot execution*/"

    def __init__(self, connection):
        self.connection = connection
        self.sql_keywords = None

    def sql_columns(self, query, query_params=None):
        """
        Get result set columns of SQL query without executing it

        Args:
            query:
                SQL query text, possibly with placeholders.
            query_params:
                Values for placeholders.

        Returns:
            Columns of SQL query result without executing it.
            Output format is similar to :meth:`pyexasol.ExaStatement.columns`.
        """
        st = self.connection.cls_statement(
            self.connection, query, query_params, prepare=True
        )
        columns = st.columns()
        st.close()

        return columns

    def schema_exists(self, schema_name):
        """
        Check if schema exists.

        Args:
            schema_name:
                Name of the schema to check.

        Returns:
            ``True`` if the schema exists, otherwise ``False``.
        """
        object_name = self.connection.format.default_format_ident_value(schema_name)

        if self.connection.protocol_version() >= constant.PROTOCOL_V2:
            st = self.execute_meta_nosql(
                "getSchemas",
                {
                    "schema": object_name,
                },
            )
        else:
            st = self.execute_snapshot(
                """
                SELECT 1
                FROM sys.exa_schemas
                WHERE schema_name={object_name}
            """,
                {
                    "object_name": object_name,
                },
            )

        return st.rowcount() > 0

    def table_exists(self, table_name):
        """
        Check if table exists.

        Args:
            table_name:
                Name of the table to check for.
                If schema was not specified, ``current_schema`` is used.

        Returns:
            ``True`` if the table exists, otherwise ``False``.
        """
        if isinstance(table_name, tuple):
            object_schema = self.connection.format.default_format_ident_value(
                table_name[0]
            )
            object_name = self.connection.format.default_format_ident_value(
                table_name[1]
            )
        else:
            object_schema = self.connection.current_schema()
            object_name = self.connection.format.default_format_ident_value(table_name)

        if self.connection.protocol_version() >= constant.PROTOCOL_V2:
            st = self.execute_meta_nosql(
                "getTables",
                {
                    "schema": object_schema,
                    "table": object_name,
                    "tableTypes": ["TABLE"],
                },
            )
        else:
            st = self.execute_snapshot(
                """
                SELECT 1
                FROM sys.exa_all_tables
                WHERE table_schema={object_schema}
                    AND table_name={object_name}
            """,
                {
                    "object_schema": object_schema,
                    "object_name": object_name,
                },
            )

        return st.rowcount() > 0

    def view_exists(self, view_name):
        """
        Check if view exists.

        Args:
            view_name:
                Name of the table to check for.
                If schema was not specified, ``current_schema`` is used.

        Returns:
            ``True`` if the view exists, otherwise ``False``.
        """
        if isinstance(view_name, tuple):
            object_schema = self.connection.format.default_format_ident_value(
                view_name[0]
            )
            object_name = self.connection.format.default_format_ident_value(
                view_name[1]
            )
        else:
            object_schema = self.connection.current_schema()
            object_name = self.connection.format.default_format_ident_value(view_name)

        if self.connection.protocol_version() >= constant.PROTOCOL_V2:
            st = self.execute_meta_nosql(
                "getTables",
                {
                    "schema": object_schema,
                    "table": object_name,
                    "tableTypes": ["VIEW"],
                },
            )
        else:
            st = self.execute_snapshot(
                """
                SELECT 1
                FROM sys.exa_all_views
                WHERE view_schema={object_schema}
                    AND view_name={object_name}
            """,
                {
                    "object_schema": object_schema,
                    "object_name": object_name,
                },
            )

        return st.rowcount() > 0

    def list_schemas(self, schema_name_pattern="%"):
        """
        List Schemas.

        Args:
            schema_name_pattern:
                Schema name or LIKE-pattern to filter on.
                (default: ``'%'``)

        Returns:
            List of schemas from `EXA_SCHEMAS <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_schemas.htm>`_ system view matching LIKE-pattern.

        Note:
            Patterns are case-sensitive. You may escape LIKE-patterns.
        """
        st = self.execute_snapshot(
            """
            SELECT *
            FROM sys.exa_schemas
            WHERE schema_name LIKE {schema_name_pattern}
            ORDER BY schema_name ASC
        """,
            {
                "schema_name_pattern": schema_name_pattern,
            },
        )

        return st.fetchall()

    def list_tables(self, table_schema_pattern="%", table_name_pattern="%"):
        """
        List Tables.

        Args:
            table_schema_pattern:
                Schema name or LIKE-pattern to filter on.
                (default: ``'%'``)
            table_name_pattern:
                Table name or LIKE-pattern to filter on.
                (default: ``'%'``)

        Returns:
            List of tables from `EXA_ALL_TABLES <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_all_tables.htm>`_ system view matching LIKE-pattern.

        Note:
            Patterns are case-sensitive. You may escape LIKE-patterns.
        """
        st = self.execute_snapshot(
            """
            SELECT *
            FROM sys.exa_all_tables
            WHERE table_schema LIKE {table_schema_pattern}
                AND table_name LIKE {table_name_pattern}
            ORDER BY table_schema ASC, table_name ASC
        """,
            {
                "table_schema_pattern": table_schema_pattern,
                "table_name_pattern": table_name_pattern,
            },
        )

        return st.fetchall()

    def list_views(self, view_schema_pattern="%", view_name_pattern="%"):
        """
        List Views.

        Args:
            view_schema_pattern:
                Schema name or LIKE-pattern to filter on.
                (default: ``'%'``)
            view_name_pattern:
                Table name or LIKE-pattern to filter on.
                (default: ``'%'``)

        Returns:
            List of views from `EXA_ALL_VIEWS <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_all_views.htm>`_ system view matching LIKE-pattern.

        Note:
            Patterns are case-sensitive. You may escape LIKE-patterns.
        """
        st = self.execute_snapshot(
            """
            SELECT *
            FROM sys.exa_all_views
            WHERE view_schema LIKE {view_schema_pattern}
                AND view_name LIKE {view_name_pattern}
            ORDER BY view_schema ASC, view_name ASC
        """,
            {
                "view_schema_pattern": view_schema_pattern,
                "view_name_pattern": view_name_pattern,
            },
        )

        return st.fetchall()

    def list_columns(
        self,
        column_schema_pattern="%",
        column_table_pattern="%",
        column_object_type_pattern="%",
        column_name_pattern="%",
    ):
        """
        List Columns.

        Args:
            column_schema_pattern:
                Schema name or LIKE-pattern to filter on.
                (default: ``'%'``)
            column_table_pattern:
                Table name or LIKE-pattern to filter on.
                (default: ``'%'``)
            column_object_type_pattern:
                Object type or LIKE-pattern to filter on.
                (default: ``'%'``)
            column_name_pattern:
                Column name or LIKE-pattern to filter on.
                (default: ``'%'``)

        Returns:
            List of columns from `EXA_ALL_COLUMNS <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_all_columns.htm>`_ system view matching LIKE-pattern.

        Note:
            Patterns are case-sensitive. You may escape LIKE-patterns.
        """
        st = self.execute_snapshot(
            """
            SELECT *
            FROM sys.exa_all_columns
            WHERE column_schema LIKE {column_schema_pattern}
                AND column_table LIKE {column_table_pattern}
                AND column_object_type LIKE {column_object_type_pattern}
                AND column_name LIKE {column_name_pattern}
        """,
            {
                "column_schema_pattern": column_schema_pattern,
                "column_table_pattern": column_table_pattern,
                "column_object_type_pattern": column_object_type_pattern,
                "column_name_pattern": column_name_pattern,
            },
        )

        return st.fetchall()

    def list_objects(
        self,
        object_name_pattern="%",
        object_type_pattern="%",
        owner_pattern="%",
        root_name_pattern="%",
    ):
        """
        List Objects.

        Args:
            object_name_pattern:
                Object name or LIKE-pattern to filter on.
                (default: ``'%'``)
            object_type_pattern:
                Object type or LIKE-pattern to filter on.
                (default: ``'%'``)
            owner_pattern:
                Owner name or LIKE-pattern to filter on.
                (default: ``'%'``)
            root_name_pattern:
                Root name or LIKE-pattern to filter on.j
                It normally refers to schema name.
                (default: ``'%'``)

        Returns:
            List of objects from `EXA_ALL_OBJECTS <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_all_objects.htm>`_ system view matching LIKE-pattern.
        Note:
            Patterns are case-sensitive. You may escape LIKE-patterns.
        """
        st = self.execute_snapshot(
            """
            SELECT *
            FROM sys.exa_all_objects
            WHERE object_name LIKE {object_name_pattern}
                AND object_type LIKE {object_type_pattern}
                AND owner LIKE {owner_pattern}
                AND root_name LIKE {root_name_pattern}
        """,
            {
                "object_name_pattern": object_name_pattern,
                "object_type_pattern": object_type_pattern,
                "owner_pattern": owner_pattern,
                "root_name_pattern": root_name_pattern,
            },
        )

        return st.fetchall()

    def list_object_sizes(
        self,
        object_name_pattern="%",
        object_type_pattern="%",
        owner_pattern="%",
        root_name_pattern="%",
    ):
        """
        List Objects with their respective size.

        Args:
            object_name_pattern:
                Object name or LIKE-pattern to filter on.
                (default: ``'%'``)
            object_type_pattern:
                Object type or LIKE-pattern to filter on.
                (default: ``'%'``)
            owner_pattern:
                Owner name or LIKE-pattern to filter on.
                (default: ``'%'``)
            root_name_pattern:
                Root name or LIKE-pattern to filter on.j
                It normally refers to schema name.
                (default: ``'%'``)

        Returns:
            List of objects with sizes from `EXA_ALL_OBJECT_SIZES <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_all_object_sizes.htm>`_ system view matching LIKE-pattern.
        Note:
            Patterns are case-sensitive. You may escape LIKE-patterns.
        """
        st = self.execute_snapshot(
            """
            SELECT *
            FROM sys.exa_all_object_sizes
            WHERE object_name LIKE {object_name_pattern}
                AND object_type LIKE {object_type_pattern}
                AND owner LIKE {owner_pattern}
                AND root_name LIKE {root_name_pattern}
        """,
            {
                "object_name_pattern": object_name_pattern,
                "object_type_pattern": object_type_pattern,
                "owner_pattern": owner_pattern,
                "root_name_pattern": root_name_pattern,
            },
        )

        return st.fetchall()

    def list_indices(
        self, index_schema_pattern="%", index_table_pattern="%", index_owner_pattern="%"
    ):
        """
        List indicies.

        Args:
            index_schema_pattern:
                Schema name or LIKE-pattern to filter on.
                (default: ``'%'``)
            index_table_pattern:
                Table name or LIKE-pattern to filter on.
                (default: ``'%'``)
            index_owner_pattern:
                Owner name or LIKE-pattern to filter on.
                (default: ``'%'``)

        Returns:
            List of indices with sizes from `EXA_ALL_INDICES <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_all_indices.htm>`_ system view matching LIKE-pattern.

        Note:
            Patterns are case-sensitive. You may escape LIKE-patterns.
        """
        st = self.execute_snapshot(
            """
            SELECT *
            FROM sys.exa_all_indices
            WHERE index_schema LIKE {index_schema_pattern}
                AND index_table LIKE {index_table_pattern}
                AND index_owner LIKE {index_owner_pattern}
        """,
            {
                "index_schema_pattern": index_schema_pattern,
                "index_table_pattern": index_table_pattern,
                "index_owner_pattern": index_owner_pattern,
            },
        )

        return st.fetchall()

    def list_sql_keywords(self):
        """
        Get reserved SQL keywords which cannot be used as identifiers without double-quote escaping

        Returns:
            List of SQL keywords from `EXA_SQL_KEYWORDS <https://docs.exasol.com/db/latest/sql_references/system_tables/metadata/exa_sql_keywords.htm>`_ system view.

        Warning:
            Never hardcode this list! It might change with next Exasol server version without warning
        Note:
            These keywords cannot be used as identifiers without double quotes.
        """
        if not self.sql_keywords:
            if self.connection.protocol_version() >= constant.PROTOCOL_V2:
                st = self.execute_meta_nosql("getKeywords")

                self.sql_keywords = [
                    r["KEYWORD"] for r in st.fetchall() if r["RESERVED"] is True
                ]
            else:
                st = self.execute_snapshot(
                    """
                    SELECT keyword
                    FROM EXA_SQL_KEYWORDS
                    WHERE reserved IS TRUE
                    ORDER BY keyword
                """
                )

                self.sql_keywords = st.fetchcol()

        return self.sql_keywords

    def execute_snapshot(self, query, query_params=None):
        """
        Execute query in snapshot transaction mode using SQL hint

        Args:
            query:
                SQL query text, possibly with placeholders.
            query_params:
                Values for placeholders.
        Returns:
            :class:`ExaStatement`

        Warning:
            Please do not try to query normal tables with this method. It will fail during creation of indices or statistics objects.

        Note:
            ``fetch_dict=Tru`` is enforced to prevent users from relying on order of columns in system views
        """
        options = {
            "fetch_dict": True,
        }

        return self.connection.cls_statement(
            self.connection,
            f"{self.snapshot_execution_hint}{query}",
            query_params,
            **options,
        )

    def execute_meta_nosql(self, meta_command, meta_params=None):
        """
        Execute no SQL metadata command introduced in Exasol 7.0+

        Args:
            meta_command:
                Metadata command.
            meta_params:
                Parameters for metadata command.

        Returns:
            :class:`ExaStatement`

        Note:
            This feature requires WebSocket protocol v2 or higher

            List of available commands can be found `here <https://github.com/exasol/websocket-api/blob/master/docs/WebsocketAPIV2.md#metadata-related-commands>`_.
        """
        if self.connection.protocol_version() < constant.PROTOCOL_V2:
            raise ExaRuntimeError(
                self.connection,
                "Protocol version 2 is required to execute nosql metadata commands",
            )

        # Security check, prevents execution of dangerous commands if meta_command argument is dynamic
        if meta_command[0:3] != "get":
            raise ExaRuntimeError(
                self.connection, "Meta command name should start with prefix 'get*'"
            )

        options = {
            "fetch_dict": True,
        }

        return self.connection.cls_statement(
            self.connection, meta_command, meta_params, meta_nosql=True, **options
        )

    def __repr__(self):
        return f"<{self.__class__.__name__} session_id={self.connection.session_id()}>"



"""
Simple benchmark for Exasol

"""

import multiprocessing
import os
import pathlib
import queue
import sys
import time

import pyexasol

BUSY_LOOP_DELAY = 0.2
STARTUP_WORKER_DELAY = 1


class ExaBenchWorker(multiprocessing.Process):
    def __init__(self, proc_num, args, sql_text, queue):
        self.proc_num = proc_num
        self.args = args
        self.sql_text = sql_text
        self.queue = queue

        super().__init__()

    def run(self):
        C = pyexasol.connect(
            dsn=args.dsn,
            user=args.user,
            password=args.password,
            schema=args.schema,
            autocommit=False,
        )
        self.queue.put(C.session_id())

        sql_comment = f"/* bench_run_id={self.proc_num} */"

        try:
            C.execute(f"ALTER SESSION SET QUERY_CACHE='OFF' {sql_comment}")

            for sql_query in sql_text.split(";"):
                formatted_sql_query = sql_query.strip("\n ")
                if len(formatted_sql_query) == 0:
                    continue

                C.execute(f"{formatted_sql_query} {sql_comment}")

            C.execute(f"COMMIT {sql_comment}")
        except pyexasol.ExaQueryError as e:
            self.queue.put(str(e))
        finally:
            C.close()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(prog=f"python -m pyexasol_utils.bench [action]")

    parser.add_argument(
        "action",
        help="Benchmark action (prepare, set1, set2, etc.), OR path to SQL file with commands",
    )
    parser.add_argument(
        "--dsn", default=os.environ.get("EXAHOST", ""), help="Exasol DSN (host:port)"
    )
    parser.add_argument(
        "--user", default=os.environ.get("EXAUID", ""), help="Exasol user name"
    )
    parser.add_argument(
        "--password", default=os.environ.get("EXAPWD", ""), help="Exasol password"
    )
    parser.add_argument(
        "--schema",
        default=os.environ.get("EXASCHEMA", "PYEXASOL_BENCH"),
        help="Exasol schema for tests (default: PYEXASOL_BENCH)",
    )
    parser.add_argument(
        "--scale",
        default=1.0,
        type=float,
        help="Scaling factor for the amount of generated data (default: 1.0)",
    )
    parser.add_argument(
        "--parallel", default=1, type=int, help="Number of parallel executions"
    )

    args = parser.parse_args()

    # Check connectivity first
    C = pyexasol.connect(dsn=args.dsn, user=args.user, password=args.password)

    # Prepare schema
    if args.action == "prepare":
        if args.parallel > 1:
            raise ValueError(
                "Cannot run [prepare.sql] with more than 1 parallel processes"
            )

        if not C.meta.schema_exists(args.schema):
            C.execute("CREATE SCHEMA {schema!i}", {"schema": args.schema})

    # Get SQL text
    bench_sql_path = pathlib.Path(__file__).parent / "bench_sql" / f"{args.action}.sql"
    bench_sql_path.resolve()

    if not bench_sql_path.is_file():
        bench_sql_path = pathlib.Path(args.action)

        if not bench_sql_path.is_file():
            raise ValueError(
                "Action argument is neither a standard action, nor path to SQL file"
            )

    sql_text = C.format.format(bench_sql_path.read_text(), scale=args.scale)

    # Show startup message
    if args.action == "prepare":
        print(
            f"Running [{bench_sql_path.name}] with scale factor of [{args.scale}]",
            file=sys.stderr,
        )
    else:
        print(
            f"Running [{bench_sql_path.name}] in [{args.parallel}] parallel processes",
            file=sys.stderr,
        )

    # Start worker processes
    process_pool: list = list()
    message_queue: multiprocessing.Queue = multiprocessing.Queue()

    for i in range(args.parallel):
        proc = ExaBenchWorker(i, args, sql_text, message_queue)
        proc.start()

        process_pool.append(proc)
        time.sleep(STARTUP_WORKER_DELAY)

    while len(process_pool) > 0:
        time.sleep(BUSY_LOOP_DELAY)

        for p in process_pool:
            if not p.is_alive():
                p.join()
                process_pool.remove(p)

                print(f"Finished worker {p.proc_num}", file=sys.stderr)

    # Display errors
    error_cnt = 0
    session_id_list = list()

    while True:
        try:
            msg = str(message_queue.get_nowait())

            if msg.isnumeric():
                session_id_list.append(msg)
            else:
                print(msg, file=sys.stderr)
                error_cnt = error_cnt + 1
        except queue.Empty:
            break

    # Flush statistics
    C.execute("FLUSH STATISTICS")

    if error_cnt > 0:
        print(f"Error count: {error_cnt}", file=sys.stderr)
        exit(0)

    # Output SQL to get results
    run_id_regex_replace = (
        r"REGEXP_REPLACE(sql_text, '(?s)^.*\/\* bench_run_id=(\d+) \*\/.*$', '\1')"
    )

    print(
        f"""
Durations SQL:
SELECT session_id
    , stmt_id
    , CAST({run_id_regex_replace} AS DECIMAL(9,0)) AS run_id
    , command_name
    , duration
    , avg(duration) OVER (PARTITION BY stmt_id) AS avg_duration
    , sql_text
FROM "$EXA_AUDIT_SQL"
WHERE session_id IN ({','.join(session_id_list)})
    AND stmt_id > 1
ORDER BY 2,3;
    """,
        file=sys.stderr,
    )

    print(
        f"""
Profiling details SQL:
SELECT session_id
    , stmt_id
    , command_name
    , part_id
    , iproc
    , duration
    , part_name
    , part_info
    , object_schema
    , object_name
    , object_rows
    , in_rows
    , out_rows
    , remarks
FROM "$EXA_PROFILE_DETAILS_LAST_DAY"
WHERE session_id = {session_id_list[-1]}
    AND stmt_id > 1
ORDER BY 2,4,5;
    """,
        file=sys.stderr,
    )

    C.close()



"""
This module provides `PEP-249`_ a DBAPI compliant Connection and Cursor protocol definition.
(see also `PEP-249-connection`_ and `PEP-249-cursor`_)

.. _PEP-249-connection: https://peps.python.org/pep-0249/#connection-objects
.. _PEP-249-cursor: https://peps.python.org/pep-0249/#cursor-objects
"""

from typing import Protocol


class Connection(Protocol):
    """
    Defines a connection protocol based on `connection-objects`_.

    .. connection-objects: https://peps.python.org/pep-0249/#connection-objects
    """

    def connect(self):
        """
        Connect to the database.

        Attention:
            Addition not required by PEP-249.
        """

    def close(self):
        """
        Close the connection now (rather than whenever .__del__() is called).

        The connection will be unusable from this point forward; an Error (or subclass)
        exception will be raised if any operation is attempted with the connection.
        The same applies to all cursor objects trying to use the connection.
        Note that closing a connection without committing the changes first will cause
        an implicit rollback to be performed.
        """

    def commit(self):
        """
        Commit any pending transaction to the database.

        Note:
            If the database supports an auto-commit feature, this must be initially off.
            An interface method may be provided to turn it back on. Database modules
            that do not support transactions should implement this method with
            void functionality.
        """

    def rollback(self):
        """
        This method is optional since not all databases provide transaction support.

        In case a database does provide transactions this method causes the database
        to roll back to the start of any pending transaction. Closing a connection
        without committing the changes first will cause an implicit rollback
        to be performed.
        """

    def cursor(self):
        """
        Return a new Cursor Object using the connection.

        If the database does not provide a direct cursor concept, the module will have
        to emulate cursors using other means to the extent needed
        by this specification.
        """


class Cursor(Protocol):
    """
    Defines a protocol which is compliant with `cursor-objects`_.

    .. cursor-objects: https://peps.python.org/pep-0249/#cursor-objects
    """

    @property
    def arraysize(self):
        """
        This read/write attribute specifies the number of rows to fetch
        at a time with .fetchmany().

        It defaults to 1, meaning to fetch a single row at a time.
        Implementations must observe this value with respect to the .fetchmany() method,
        but are free to interact with the database a single row at a time.
        It may also be used in the implementation of .executemany().
        """

    @property
    def description(self):
        """
        This read-only attribute is a sequence of 7-item sequences.

        Each of these sequences contains information describing one result column:

            * name
            * type_code
            * display_size
            * internal_size
            * precision
            * scale
            * null_ok

        The first two items (name and type_code) are mandatory, the other five
        are optional and are set to None if no meaningful values can be provided.

        This attribute will be None for operations that do not return rows or if
        the cursor has not had an operation invoked via the .execute*() method yet.
        """

    @property
    def rowcount(self):
        """
        This read-only attribute specifies the number of rows that the last .execute*()
        produced (for DQL statements like SELECT) or affected (for DML statements
        like UPDATE or INSERT).

        The attribute is -1 in case no .execute*() has been performed on the cursor or
        the rowcount of the last operation cannot be determined by the interface.

        .. note::

            Future versions of the DB API specification could redefine the latter case
            to have the object return None instead of -1.
        """

    def callproc(self, procname, parameters):
        """
        Call a stored database procedure with the given name.
        (This method is optional since not all databases provide stored procedures)

        The sequence of parameters must contain one entry for each argument that the
        procedure expects. The result of the call is returned as a modified copy of
        the input sequence. Input parameters are left untouched, output and
        input/output parameters replaced with possibly new values.

        The procedure may also provide a result set as output. This must then be
        made available through the standard .fetch*() methods.
        """

    def close(self):
        """
        Close the cursor now (rather than whenever __del__ is called).

        The cursor will be unusable from this point forward; an Error (or subclass)
        exception will be raised if any operation is attempted with the cursor.
        """

    def execute(self, operation, parameters=None):
        """
        Prepare and execute a database operation (query or command).

        Parameters may be provided as sequence or mapping and will be bound to
        variables in the operation. Variables are specified in a database-specific
        notation (see the module’s paramstyle attribute for details).

        A reference to the operation will be retained by the cursor.
        If the same operation object is passed in again, then the cursor can optimize
        its behavior. This is most effective for algorithms where the same operation
        is used, but different parameters are bound to it (many times).

        For maximum efficiency when reusing an operation, it is best to use the
        .setinputsizes() method to specify the parameter types and sizes ahead of time.
        It is legal for a parameter to not match the predefined information;
        the implementation should compensate, possibly with a loss of efficiency.

        The parameters may also be specified as list of tuples to e.g. insert multiple
        rows in a single operation, but this kind of usage is deprecated: .executemany()
        should be used instead.

        Return values are not defined.
        """

    def executemany(self, operation, seq_of_parameters):
        """
        Prepare a database operation (query or command) and then execute it against all
        parameter sequences or mappings found in the sequence seq_of_parameters.

        Modules are free to implement this method using multiple calls to the .execute()
        method or by using array operations to have the database process the sequence
        as a whole in one call.

        Use of this method for an operation which produces one or more result sets
        constitutes undefined behavior, and the implementation is permitted
        (but not required) to raise an exception when it detects that a result set
        has been created by an invocation of the operation.

        The same comments as for .execute() also apply accordingly to this method.

        Return values are not defined.
        """

    def fetchone(self):
        """
        Fetch the next row of a query result set, returning a single sequence, or None
        when no more data is available.

        An Error (or subclass) exception is raised if the previous call to .execute*()
        did not produce any result set or no call was issued yet.
        """

    def fetchmany(self, size=None):
        """
        Fetch the next set of rows of a query result, returning a sequence of sequences
        (e.g. a list of tuples).

        An empty sequence is returned when no more rows are available. The number of
        rows to fetch per call is specified by the parameter. If it is not given,
        the cursor’s arraysize determines the number of rows to be fetched. The method
        should try to fetch as many rows as indicated by the size parameter.
        If this is not possible due to the specified number of rows not being available,
        fewer rows may be returned.

        An Error (or subclass) exception is raised if the previous call to .execute*()
        did not produce any result set or no call was issued yet.

        Note there are performance considerations involved with the size parameter.
        For optimal performance, it is usually best to use the .arraysize attribute.
        If the size parameter is used, then it is best for it to retain the same value
        from one .fetchmany() call to the next.
        """

    def fetchall(self):
        """
        Fetch all (remaining) rows of a query result, returning them as a sequence of
        sequences (e.g. a list of tuples).

        Note that the cursor’s arraysize attribute can affect the performance of this
        operation. An Error (or subclass) exception is raised if the previous call to
        .execute*() did not produce any result set or no call was issued yet.
        """

    def nextset(self):
        """
        This method will make the cursor skip to the next available set, discarding any
        remaining rows from the current set.
        (This method is optional since not all databases support multiple result sets)

        If there are no more sets, the method returns None. Otherwise, it returns a true
        value and subsequent calls to the .fetch*() methods will return rows from the
        next result set.

        An Error (or subclass) exception is raised if the previous call to .execute*()
        did not produce any result set or no call was issued yet.
        """

    def setinputsizes(self, sizes):
        """
        This can be used before a call to .execute*() to predefine memory areas for the
        operation’s parameters.

        sizes is specified as a sequence — one item for each input parameter. The item
        should be a Type Object that corresponds to the input that will be used, or it
        should be an integer specifying the maximum length of a string parameter. If the
        item is None, then no predefined memory area will be reserved for that column
        (this is useful to avoid predefined areas for large inputs).

        This method would be used before the .execute*() method is invoked.

        Implementations are free to have this method do nothing and users are free
        to not use it.
        """

    def setoutputsizes(self, size, column):
        """
        Set a column buffer size for fetches of large columns (e.g. LONGs, BLOBs, etc.).

        The column is specified as an index into the result sequence. Not specifying
        the column will set the default size for all large columns in the cursor.
        This method would be used before the .execute*() method is invoked.

        Implementations are free to have this method do nothing and users are free
        to not use it.
        """
